"""
_exploration/__init__.py
=========================

.. module:: __init__
   :platform: Unix
   :synopsis: Initialization module for the _exploration package.

Module Overview
---------------

This module initializes the exploration functions of the `hsi-wizard`.

:no-index:
"""

from .cluster import isodata
