"""
_utils/__init__.py
===================

.. module:: __init__
   :platform: Unix
   :synopsis: Initialization module for the hsi-wizard package.

Module Overview
---------------

This module initializes the _utils package, making utility functions
available for use throughout the hsi-wizard package.

Functions
---------

.. autofunction:: find_nex_greater_wave
.. autofunction:: find_nex_smaller_wave

"""
from .helper import find_nex_greater_wave, find_nex_smaller_wave
